/*
 * Library interface to Linux x86 Performance-Monitoring Counters driver.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */
#include "asm/perfctr.h"
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include "libperfctr.h"

static int perfctr_fd = -1;	/* to force initialization */

#define CHECK_INIT() \
do { \
  if( perfctr_fd < 0 && perfctr_init() < 0 ) \
    return -1; \
} while(0)

static int have_direct_access;
static struct perfctr_control current_control;
static struct perfctr_counters base; /* updated by kernel */

#define rdtsc(low,high) __asm__ __volatile__("rdtsc" : "=a"(low), "=d"(high))
#define rdpmc(ctr,low,high) __asm__ __volatile__("rdpmc" : "=a"(low), "=d"(high) : "c"(ctr))

static int perfctr_init(void)
{
    struct perfctr_info info;

    if( (perfctr_fd = open("/dev/perfctr", O_RDONLY)) < 0 )
	return -1;
    if( ioctl(perfctr_fd, PERFCTR_INFO, &info) < 0 )
	return -1;
    /* This will be false for pre-MMX P5 (no rdpmc) and
     * WinChip processors (enabling perfctrs disables rdtsc).
     */
    have_direct_access = info.have_rdpmc && info.have_rdtsc;
    return 0;
}

int perfctr_info(struct perfctr_info *info)
{
    CHECK_INIT();
    return ioctl(perfctr_fd, PERFCTR_INFO, info);
}

int perfctr_enable(struct perfctr_control *control)
{
    struct perfctr_enable enable;

    CHECK_INIT();
    current_control = *control;
    enable.control = *control;
    enable.user = &base;
    if( ioctl(perfctr_fd, PERFCTR_ENABLE, &enable) < 0 ) {
	current_control.evntsel[0] = 0;
	return -1;
    }
    return 0;
}

int perfctr_read(struct perfctr_state *state)
{
    CHECK_INIT();
    if( have_direct_access && current_control.evntsel[0] ) {
	perfctr_uint64_t tsc_start, tsc, pmc[2];
	do {
	    tsc_start = base.tsc_start;
	    rdtsc(tsc.u32[0], tsc.u32[1]);
	    rdpmc(0, pmc[0].u32[0], pmc[0].u32[1]);
	    rdpmc(1, pmc[1].u32[0], pmc[1].u32[1]);
	    state->counters.tsc.u64 = base.tsc.u64 + (tsc.u64 - tsc_start.u64);
	    state->counters.pmc[0].u64 = base.pmc[0].u64 + pmc[0].u64;
	    state->counters.pmc[1].u64 = base.pmc[1].u64 + pmc[1].u64;
	} while( tsc_start.u32[0] != base.tsc_start.u32[0] );
	state->counters.tsc_start = tsc_start;
	state->control = current_control;
	state->user = &base;
	return 0;
    } else
	return ioctl(perfctr_fd, PERFCTR_READ, state);
}
